#include "database.h"

#include <assert.h>
#include <malloc.h>
#include <stdbool.h>
#include <string.h>
#include <sqlite3.h>

static JSClassID _database_class_id;
static int _database_count;

typedef struct _database_t {
	JSContext* context;
	JSValue object;
	void* task;
	sqlite3* db;
	const char* id;
} database_t;

static JSValue _database_create(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv, int magic, JSValue* data);
static void _database_finalizer(JSRuntime *runtime, JSValue value);

static JSValue _database_get(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv);
static JSValue _database_set(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv);
static JSValue _database_remove(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv);
static JSValue _database_get_all(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv);

void tf_database_init(JSContext* context, sqlite3* sqlite) {
	JS_NewClassID(&_database_class_id);
	JSClassDef def = {
		.class_name = "Database",
		.finalizer = &_database_finalizer,
	};
	if (JS_NewClass(JS_GetRuntime(context), _database_class_id, &def) != 0) {
		printf("Failed to register database.\n");
	}

	JSValue global = JS_GetGlobalObject(context);
	JSValue data[] = { JS_NewInt64(context, (int64_t)(intptr_t)sqlite) };
	JSValue constructor = JS_NewCFunctionData(context, _database_create, 0, 0, 1, data);
	JS_SetConstructorBit(context, constructor, true);
	JS_SetPropertyStr(context, global, "Database", constructor);
	JS_FreeValue(context, global);
}

static JSValue _database_create(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv, int magic, JSValue* data) {
	++_database_count;
	JSValue object = JS_NewObjectClass(context, _database_class_id);
	sqlite3* db = NULL;
	JS_ToInt64(context, (int64_t*)&db, data[0]);

	database_t* database = malloc(sizeof(database_t));
	*database = (database_t) {
		.task = JS_GetContextOpaque(context),
		.context = context,
		.object = object,
		.db = db,
	};
	const char* id = JS_ToCString(context, argv[0]);
	database->id = strdup(id);
	JS_FreeCString(context, id);
	JS_SetOpaque(object, database);

	JS_SetPropertyStr(context, object, "get", JS_NewCFunction(context, _database_get, "get", 1));
	JS_SetPropertyStr(context, object, "set", JS_NewCFunction(context, _database_set, "set", 2));
	JS_SetPropertyStr(context, object, "remove", JS_NewCFunction(context, _database_remove, "remove", 1));
	JS_SetPropertyStr(context, object, "getAll", JS_NewCFunction(context, _database_get_all, "getAll", 0));

	return object;
}

static void _database_finalizer(JSRuntime *runtime, JSValue value) {
	database_t* database = JS_GetOpaque(value, _database_class_id);
	if (database) {
		free((void*)database->id);
		free(database);
	}
	--_database_count;
}

static JSValue _database_get(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv) {
	JSValue entry = JS_UNDEFINED;
	database_t* database = JS_GetOpaque(this_val, _database_class_id);
	if (database) {
		sqlite3_stmt* statement;
		if (sqlite3_prepare(database->db, "SELECT value FROM properties WHERE id = $1 AND key = $2", -1, &statement, NULL) == SQLITE_OK) {
			size_t length;
			const char* keyString = JS_ToCStringLen(context, &length, argv[0]);
			if (sqlite3_bind_text(statement, 1, database->id, -1, NULL) == SQLITE_OK &&
				sqlite3_bind_text(statement, 2, keyString, length, NULL) == SQLITE_OK &&
				sqlite3_step(statement) == SQLITE_ROW) {
				entry = JS_NewStringLen(context, (const char*)sqlite3_column_text(statement, 0), sqlite3_column_bytes(statement, 0));
			}
			sqlite3_finalize(statement);
		}
	}
	return entry;
}

JSValue _database_set(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv) {
	database_t* database = JS_GetOpaque(this_val, _database_class_id);
	if (database) {
		sqlite3_stmt* statement;
		if (sqlite3_prepare(database->db, "INSERT OR REPLACE INTO properties (id, key, value) VALUES ($1, $2, $3)", -1, &statement, NULL) == SQLITE_OK) {
			size_t keyLength;
			const char* keyString = JS_ToCStringLen(context, &keyLength, argv[0]);
			size_t valueLength;
			const char* valueString = JS_ToCStringLen(context, &valueLength, argv[1]);
			if (sqlite3_bind_text(statement, 1, database->id, -1, NULL) == SQLITE_OK &&
				sqlite3_bind_text(statement, 2, keyString, keyLength, NULL) == SQLITE_OK &&
				sqlite3_bind_text(statement, 3, valueString, valueLength, NULL) == SQLITE_OK &&
				sqlite3_step(statement) == SQLITE_OK) {
			}
			sqlite3_finalize(statement);
		}
	}
	return JS_UNDEFINED;
}

JSValue _database_remove(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv) {
	database_t* database = JS_GetOpaque(this_val, _database_class_id);
	if (database) {
		sqlite3_stmt* statement;
		if (sqlite3_prepare(database->db, "DELETE FROM properties WHERE id = $1 AND key = $2", -1, &statement, NULL) == SQLITE_OK) {
			size_t keyLength;
			const char* keyString = JS_ToCStringLen(context, &keyLength, argv[0]);
			if (sqlite3_bind_text(statement, 1, database->id, -1, NULL) == SQLITE_OK &&
				sqlite3_bind_text(statement, 2, keyString, keyLength, NULL) == SQLITE_OK &&
				sqlite3_step(statement) == SQLITE_OK) {
			}
			sqlite3_finalize(statement);
		}
	}
	return JS_UNDEFINED;
}

JSValue _database_get_all(JSContext* context, JSValueConst this_val, int argc, JSValueConst* argv) {
	JSValue array = JS_UNDEFINED;
	database_t* database = JS_GetOpaque(this_val, _database_class_id);
	if (database) {
		sqlite3_stmt* statement;
		if (sqlite3_prepare(database->db, "SELECT key, value FROM properties WHERE id = $1", -1, &statement, NULL) == SQLITE_OK) {
			if (sqlite3_bind_text(statement, 1, database->id, -1, NULL) == SQLITE_OK) {
				array = JS_NewArray(context);
				uint32_t index = 0;
				while (sqlite3_step(statement) == SQLITE_ROW) {
					JS_SetPropertyUint32(context, array, index++, JS_NewStringLen(context, (const char*)sqlite3_column_text(statement, 0), sqlite3_column_bytes(statement, 0)));
				}
			}
			sqlite3_finalize(statement);
		}
	}
	return array;
}
